from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, TypeGuard

from .codec import (AudioCodecParameters, AudioEncodingParameters,
                    VideoCodecParameters, VideoEncodingParameters)
from .common import EventDispatcher
from .mediaformat import Resolution, TrackKink
from .transport import TransportConstraints, TransportSettings
from .webrtc import RTCRtpEncodingParameters


@dataclass
class AudioPublicationSettings:
    codec: AudioCodecParameters | None = None
    _trackId: str | None = None
    
@dataclass
class VideoPublicationSettings:
    codec: VideoCodecParameters | None = None
    resolution: Resolution | None = None
    frameRate: float | None = None
    bitrate: float | str | None = None
    keyFrameInterval: float | None = None
    rid: str | None = None
    _trackId: str | None = None

@dataclass
class PublicationSettings:
    audio: list[AudioPublicationSettings] = field(default_factory=list)
    video: list[VideoPublicationSettings] = field(default_factory=list)

class Publication(EventDispatcher):
    id: str
    transport: TransportSettings
    stop: Callable[[], Awaitable[Any]]
    mute: Callable[[TrackKink], Awaitable[Any]]
    unmute: Callable[[TrackKink], Awaitable[Any]]
    
    def __init__(
        self, id: str,
        transport: TransportSettings,
        stop: Callable[[], Any],
        mute: Callable[[TrackKink], Any],
        unmute: Callable[[TrackKink], Any],
    ) -> None:
        super().__init__()
        self.id = id
        self.transport = transport
        self.stop = stop
        self.mute = mute
        self.unmute = unmute
   
@dataclass
class PublishOptions:
    """
    PublishOptions defines options for publishing a LocalStream.
    """
    
    audio: list[AudioEncodingParameters] | list[RTCRtpEncodingParameters] | bool | None = None
    """
    Parameters for audio RtpSender.
    """
    video: list[VideoEncodingParameters] | list[RTCRtpEncodingParameters] | bool | None = None
    """
    Parameters for video RtpSender.
    """
    transport: TransportConstraints | None = None
    
    def __post_init__(self):
        if (self.isAudioEncodingParameters(self.audio) and self.isRTCRtpEncodingParameters(self.video)) or (self.isVideoEncodingParameters(self.video) and self.isRTCRtpEncodingParameters(self.audio)):
            raise TypeError('Mixing RTCRtpEncodingParameters and AudioEncodingParameters/VideoEncodingParameters is not allowed.')
        
    @staticmethod
    def isAudioEncodingParameters(value: list[AudioEncodingParameters] | list[RTCRtpEncodingParameters] | bool | None) -> TypeGuard[list[AudioEncodingParameters]]:
        if isinstance(value, list):
            if len(value) > 0:
                return isinstance(value[0], AudioEncodingParameters)
        return False
    
    @staticmethod
    def isVideoEncodingParameters(value: list[VideoEncodingParameters] | list[RTCRtpEncodingParameters] | bool | None) -> TypeGuard[list[VideoEncodingParameters]]:
        if isinstance(value, list):
            if len(value) > 0:
                return isinstance(value[0], VideoEncodingParameters)
        return False
    
    @staticmethod
    def isRTCRtpEncodingParameters(value: list[AudioEncodingParameters] | list[VideoEncodingParameters] | list[RTCRtpEncodingParameters] | bool | None) -> TypeGuard[list[RTCRtpEncodingParameters]]:
        if isinstance(value, list):
            if len(value) > 0:
                return isinstance(value[0], RTCRtpEncodingParameters)
        return False
    


from dataclasses import dataclass
from typing import Literal

from .common import Transceiver

TransportType = Literal['quic', 'webrtc']

@dataclass
class TransportConstraints:
    type: TransportType
    """
    Transport type for publication and subscription.
    """
    id: str | None = None
    """
    Transport ID. Undefined transport ID results server to assign a new
    one. It should always be undefined if transport type is webrtc since the
    webrtc agent of OWT server doesn't support multiple transceivers on a
    single PeerConnection.
    """

@dataclass
class TransportSettings:
    type: TransportType
    id: str
    rtpTransceivers: list[Transceiver]


from dataclasses import dataclass
import sys
from typing import cast
if sys.version_info < (3, 11):
    from typing_extensions import Literal, TypedDict
else:
    from typing import Literal, TypedDict

RTCDtxStatus = Literal['Disabled', 'Enabled']

class RTCRtpCodingParametersDict(TypedDict):
    rid: str | None
    """
    A DOMString which, if set, specifies an RTP stream ID (RID) to be sent using the RID header extension. This parameter cannot be modified using setParameters(). Its value can only be set when the transceiver is first created.
    """

@dataclass
class RTCRtpCodingParameters:
    rid: str | None = None
    """
    A DOMString which, if set, specifies an RTP stream ID (RID) to be sent using the RID header extension. This parameter cannot be modified using setParameters(). Its value can only be set when the transceiver is first created.
    """
    
    def to_dict(self) -> RTCRtpCodingParametersDict:
        return cast(RTCRtpCodingParametersDict, self.__dict__)
    
class RTCRtpEncodingParametersDict(RTCRtpCodingParametersDict):
    """
    An instance of the WebRTC API's RTCRtpEncodingParameters dictionary describes a single configuration of a codec for an RTCRtpSender. It's used in the RTCRtpSendParameters describing the configuration of an RTP sender's encodings; RTCRtpDecodingParameters is used to describe the configuration of an RTP receiver's encodings.
    """
    
    active: bool
    """
    If true, the described encoding is currently actively being used. That is, for RTP senders, the encoding is currently being used to send data, while for receivers, the encoding is being used to decode received data. The default value is true.
    """
    codecPayloadType: int | None
    """
    When describing a codec for an RTCRtpSender, codecPayloadType is a single 8-bit byte (or octet) specifying the codec to use for sending the stream; the value matches one from the owning RTCRtpParameters object's codecs parameter. This value can only be set when creating the transceiver; after that, this value is read only.
    """
    dtx: RTCDtxStatus | None
    """
    Only used for an RTCRtpSender whose kind is audio, this property indicates whether or not to use discontinuous transmission (a feature by which a phone is turned off or the microphone muted automatically in the absence of voice activity). The value is taken from the enumerated string type RTCDtxStatus.
    """
    maxBitrate: float | None
    """
    An unsigned long integer indicating the maximum number of bits per second to allow for this encoding. Other parameters may further constrain the bit rate, such as the value of maxFramerate or transport or physical network limitations.
    """
    maxFramerate: float | None
    """
    A double-precision floating-point value specifying the maximum number of frames per second to allow for this encoding.
    """
    ptime: int | None
    """
    An unsigned long integer value indicating the preferred duration of a media packet in milliseconds. This is typically only relevant for audio encodings. The user agent will try to match this as well as it can, but there is no guarantee.
    """
    scaleResolutionDownBy: float | None
    """
    Only used for senders whose track's kind is video, this is a double-precision floating-point value specifying a factor by which to scale down the video during encoding. The default value, 1.0, means that the sent video's size will be the same as the original. A value of 2.0 scales the video frames down by a factor of 2 in each dimension, resulting in a video 1/4 the size of the original. The value must not be less than 1.0 (you can't use this to scale the video up).
    """

@dataclass
class RTCRtpEncodingParameters(RTCRtpCodingParameters):
    """
    An instance of the WebRTC API's RTCRtpEncodingParameters dictionary describes a single configuration of a codec for an RTCRtpSender. It's used in the RTCRtpSendParameters describing the configuration of an RTP sender's encodings; RTCRtpDecodingParameters is used to describe the configuration of an RTP receiver's encodings.
    """
    
    active: bool = True
    """
    If true, the described encoding is currently actively being used. That is, for RTP senders, the encoding is currently being used to send data, while for receivers, the encoding is being used to decode received data. The default value is true.
    """
    codecPayloadType: int | None = None
    """
    When describing a codec for an RTCRtpSender, codecPayloadType is a single 8-bit byte (or octet) specifying the codec to use for sending the stream; the value matches one from the owning RTCRtpParameters object's codecs parameter. This value can only be set when creating the transceiver; after that, this value is read only.
    """
    dtx: RTCDtxStatus | None = None
    """
    Only used for an RTCRtpSender whose kind is audio, this property indicates whether or not to use discontinuous transmission (a feature by which a phone is turned off or the microphone muted automatically in the absence of voice activity). The value is taken from the enumerated string type RTCDtxStatus.
    """
    maxBitrate: float | None = None
    """
    An unsigned long integer indicating the maximum number of bits per second to allow for this encoding. Other parameters may further constrain the bit rate, such as the value of maxFramerate or transport or physical network limitations.
    """
    maxFramerate: float | None = None
    """
    A double-precision floating-point value specifying the maximum number of frames per second to allow for this encoding.
    """
    ptime: int | None = None
    """
    An unsigned long integer value indicating the preferred duration of a media packet in milliseconds. This is typically only relevant for audio encodings. The user agent will try to match this as well as it can, but there is no guarantee.
    """
    scaleResolutionDownBy: float | None = None
    """
    Only used for senders whose track's kind is video, this is a double-precision floating-point value specifying a factor by which to scale down the video during encoding. The default value, 1.0, means that the sent video's size will be the same as the original. A value of 2.0 scales the video frames down by a factor of 2 in each dimension, resulting in a video 1/4 the size of the original. The value must not be less than 1.0 (you can't use this to scale the video up).
    """
    
    def to_dict(self) -> RTCRtpEncodingParametersDict:
        return cast(RTCRtpEncodingParametersDict, self.__dict__)


from dataclasses import dataclass

from .participant import Participant
from .stream import RemoteStream


@dataclass
class ConferenceInfo:
    id: str
    participants: list[Participant]
    remoteStreams: list[RemoteStream]
    self: Participant


from dataclasses import dataclass
from typing import Literal

from .portal_types import Resolution as SigResolution

TrackKink = Literal['audio'] | Literal['video'] | Literal['av']

@dataclass
class Resolution:
    width: int
    height: int
    
    def to_dict(self) -> SigResolution:
        return {
            'width': self.width,
            'height': self.height,
        }


import asyncio
import logging
import sys
from dataclasses import dataclass, field
from typing import cast

from aiortc import (MediaStreamTrack, RTCConfiguration, RTCPeerConnection,
                    RTCRtpTransceiver, RTCSessionDescription)

if sys.version_info < (3, 11):
    from typing_extensions import Any, Literal
else:
    from typing import Any, Literal

from .common import (AudioTransceiver, ConferenceError, ErrorEvent,
                     EventDispatcher, MessageEvent, MuteEvent, OwtEvent,
                     Transceiver, VideoParameters, VideoTransceiver)
from .constants import AudioCodec, VideoCodec
from .mediaformat import TrackKink
from .portal_types import (CandidateMessage, OfferAnswer, StreamUpdateMessage,
                           TransportProgress)
from .publication import Publication, PublishOptions
from .sdputils import reorderCodecs, setMaxBitrate
from .signaling import SioSignaling
from .stream import LocalStream, RemoteStream
from .subscription import (AudioSubscriptionConstraints, SubscribeOptions,
                           Subscription, SubscriptionUpdateOptions,
                           VideoSubscriptionConstraints)
from .transport import TransportSettings
from .utils import AsyncExceptionCollector, chain_future, remove_none_value, Postponer

logger = logging.getLogger("OWT")
logger.setLevel(logging.DEBUG)


@dataclass
class TransceiversWithId:
    id: str | None = None
    transceivers: list[Transceiver] = field(default_factory=list)

class OwtChannel(EventDispatcher):
    pc: RTCPeerConnection
    _lock: asyncio.Lock
    _signaling: SioSignaling
    _id: str | None # transportId
    _remoteId: str | None
    _internalCount: int
    _ended: bool
    _subscriptions: dict[str, Subscription] # SubscriptionId => Subscription
    _subscribeTransceivers: dict[int, TransceiversWithId] # internalId => { id, transceivers: [Transceiver] }
    _subscribePromises: dict[int, asyncio.Future[Subscription]] # internalId => Future
    _reverseIdMap: dict[str, int] # PublicationId || SubscriptionId => internalId
    _remoteMediaTracks: dict[str, list[MediaStreamTrack]] # Key is subscription ID, value is MediaStreamTrack.
    _publications: dict[str, Publication] # PublicationId => Publication
    _publishTransceivers: dict[int, TransceiversWithId] # internalId => { id, transceivers: [Transceiver] }
    _publishPromises: dict[int, asyncio.Future[Publication]] # internalId => Future
    _sdpPromise: asyncio.Future[None]
    _sdpResolvers: list[asyncio.Future[None]]
    _sdpResolverMap: dict[int, asyncio.Future[None]]
    _sdpResolveNum: int
    _loop: asyncio.AbstractEventLoop | None
    exception_collector: AsyncExceptionCollector
    
    def __init__(self, signaling: SioSignaling, rtc_configure: RTCConfiguration | None = None, loop: asyncio.AbstractEventLoop | None = None) -> None:
        super().__init__()
        self._loop = loop
        self._lock = asyncio.Lock()
        self._ended = False
        self._id = None
        self._remoteId = None
        self._internalCount = 0
        self._signaling = signaling
        self._subscriptions = {}
        self._subscribeTransceivers = {}
        self._subscribePromises = {}
        
        self._reverseIdMap = {} # PublicationId || SubscriptionId => internalId
        
        self._remoteMediaTracks = {}
        
        self._publications = {}
        self._publishTransceivers = {}
        self._publishPromises = {}
        
        self._sdpPromise = asyncio.Future(loop=self._loop)
        self._sdpPromise.set_result(None)
        self._sdpResolvers = []
        self._sdpResolverMap = {}
        self._sdpResolveNum = 0
        self.exception_collector = AsyncExceptionCollector()
        self.pc = RTCPeerConnection(configuration=rtc_configure)
        self.pc.on('track', self._onRemoteTrackAdded)
        self.pc.on('iceconnectionstatechange', self._onIceConnectionStateChange)
        self.pc.on('connectionstatechange', self._onConnectionStateChange)
        def on_signalingstatechange():
            logger.debug(f'signaling state: {self.pc.signalingState}')
        self.pc.on('signalingstatechange', on_signalingstatechange)
        def on_icegatheringstatechange():
            logger.debug(f'ice gathering state: {self.pc.iceGatheringState}')
        self.pc.on('icegatheringstatechange', on_icegatheringstatechange)
        
    async def onMessage(self, notification: str, message: Any):
        async with self.exception_collector.collect():
            async with await Postponer().postpone(self, self.close):
                if notification == 'progress':
                    progress_message = cast(TransportProgress, message)
                    if progress_message['status'] == 'soac':
                        await self._sdpHandler(progress_message['data'])
                    elif progress_message['status'] == 'ready' and 'sessionId' in progress_message:
                        # ready消息可能发两次，第一次不带sessionId，只有id，也就是transport id，用以建立连接，第二次才带sessionId
                        sessionId = progress_message['sessionId']
                        assert sessionId is not None
                        self._readyHandler(sessionId)
                    elif progress_message['status'] == 'error':
                        # 文档中完美没提到sessionId字段，但js源码里用的是sessionId字段
                        await self._errorHandler(progress_message.get('sessionId'), progress_message['data'])
                elif notification == 'stream':
                    stream_message = cast(StreamUpdateMessage, message)
                    await self._onStreamEvent(stream_message)
                else:
                    logger.warning('Unknown notification from MCU.')
            
    async def publishWithTransceivers(self, stream: LocalStream, transceivers: list[RTCRtpTransceiver]) -> Publication:
        for t in transceivers:
            if t.direction != 'sendonly':
                raise ConferenceError('RTCRtpTransceiver\'s direction must be sendonly.')
            if t.sender.track not in stream.tracks:
                raise ConferenceError('The track associated with RTCRtpSender is not included in stream.')
            if len(transceivers) > 2:
                raise ConferenceError('At most one transceiver for audio and one transceiver for video are accepted.')
        internalId = self._createInternalId()
        await self._chainSdpPromise(internalId)
        transceiverDescription = [
            (VideoTransceiver(type='video', transceiver=t, source=stream.source.video, option=[]) if t.sender.track.kind == 'video'
            else AudioTransceiver(type='audio', transceiver=t, source=stream.source.audio, option=[])) for t in transceivers]
        self._publishTransceivers[internalId] = TransceiversWithId(transceivers=transceiverDescription)
        future: asyncio.Future[Publication] = asyncio.Future(loop=self._loop)
        self._publishPromises[internalId] = future
        try:
            localDesc = await self.pc.createOffer()
            await self.pc.setLocalDescription(localDesc)
            localDesc = self.pc.localDescription
            pub_res = await self._signaling.publish({
                'media': { 'tracks': [{ 'type': t.type, 'mid': t.transceiver.mid, 'source': t.source } for t in transceiverDescription] },
                'attributes': stream.attributes,
                'transport': {
                    'id': self._id,
                    'type': 'webrtc',
                },
            })
            publicationId = pub_res['id']
            self._publishTransceivers[internalId].id = publicationId
            self._reverseIdMap[publicationId] = internalId
            if self._id and self._id != pub_res['transportId']:
                raise ConferenceError(f'Server returns conflict ID: {pub_res["transportId"]}')
            self._id = pub_res['transportId']
            await self.dispatchEventAwaitable(MessageEvent('id', message=publicationId, origin=self._remoteId))
            await self._signaling.soac(
                self._id,
                {
                    'type': cast(Literal['offer'], localDesc.type),
                    'sdp': localDesc.sdp,
                },
            )
            return await future
        except:
            await self._unpublish(internalId)
            raise
            
    async def publish(self, stream: LocalStream, options: PublishOptions | list[RTCRtpTransceiver] | None, videoCodecs: list[str] | None) -> Publication:
        if self._ended:
            raise ConferenceError('Connection closed')
        if isinstance(options, list):
            return await self.publishWithTransceivers(stream, options)
        if options is None:
            options = PublishOptions(
                audio=stream.hasAudio(),
                video=stream.hasVideo(),
            )
        if options.audio is None:
            options.audio = stream.hasAudio()
        if options.video is None:
            options.video = stream.hasVideo()
        if (options.audio and not stream.hasAudio()) or (options.video and not stream.hasVideo()):
            raise ConferenceError('options.audio/video is inconsistent with tracks presented in the MediaStream')
        if not options.audio and not options.video:
            raise ConferenceError('Cannot publish a stream without audio and video.')
        if stream.audio_num > 1:
            logger.warning('Publishing a stream with multiple audio tracks is not fully supported.')
        if stream.video_num > 1:
            logger.warning('Publishing a stream with multiple video tracks is not fully supported.')
        
        internalId = self._createInternalId()
        await self._chainSdpPromise(internalId)
        transceivers: list[Transceiver] = []
        if stream.hasVideo():
            track = stream.videoTrack
            assert track is not None
            transceiver = self.pc.addTransceiver(track, direction='sendonly')
            transceivers.append(VideoTransceiver(
                type='video',
                transceiver=transceiver,
                source=stream.source.video,
                option=options.video,
            ))
        if stream.hasAudio():
            track = stream.audioTrack
            assert track is not None
            transceiver = self.pc.addTransceiver(track, direction='sendonly')
            transceivers.append(AudioTransceiver(
                type='audio',
                transceiver=transceiver,
                source=stream.source.audio,
                option=options.audio,
            ))
        assert internalId not in self._publishTransceivers
        self._publishTransceivers[internalId] = TransceiversWithId(transceivers=transceivers)
        future: asyncio.Future[Publication] = asyncio.Future(loop=self._loop)
        self._publishPromises[internalId] = future
        publicationId = None
        try:
            localDesc = await self.pc.createOffer()
            await self.pc.setLocalDescription(localDesc)
            localDesc = self.pc.localDescription
            pub_res = await self._signaling.publish({
                'media': { 'tracks': [{ 'type': t.type, 'mid': t.transceiver.mid, 'source': t.source } for t in transceivers] },
                'attributes': stream.attributes,
                'transport': {
                    'id': self._id or '',
                    'type': 'webrtc',
                },
            })
            publicationId = pub_res['id']
            self._publishTransceivers[internalId].id = publicationId
            self._reverseIdMap[publicationId] = internalId
            if self._id and self._id != pub_res['transportId']:
                raise ConferenceError(f'Server returns conflict ID: {pub_res["transportId"]}')
            self._id = pub_res['transportId']
            await self.dispatchEventAwaitable(MessageEvent('id', message=publicationId, origin=self._remoteId))
            if options:
                for t in transceivers:
                    localDesc.sdp = self._setRtpReceiverOptions(localDesc.sdp, options, videoCodecs, t.transceiver.mid)
                    localDesc.sdp = self._setRtpSenderOptions(localDesc.sdp, options, t.transceiver.mid)
            await self._signaling.soac(
                self._id,
                {
                    'type': cast(Literal['offer'], localDesc.type),
                    'sdp': localDesc.sdp,
                },
            )
            return await future
        except:
            await self._unpublish(internalId)
            raise
    
    async def subscribe(self, stream: RemoteStream, options: SubscribeOptions | None = None) -> Subscription:
        if self._ended:
            raise asyncio.InvalidStateError('Connection closed')
        has_audio_stream = bool(stream.settings and stream.settings.audio)
        has_video_stream = bool(stream.settings and stream.settings.video)
        if not options:
            options = SubscribeOptions(audio=has_audio_stream, video=has_video_stream)
        if not options.audio:
            options.audio = has_audio_stream
        if not options.video:
            options.video = has_video_stream
        if (options.audio and not has_audio_stream) or (options.video and not has_video_stream):
            raise ValueError('options.audio/video cannot be true or an object if there is no audio/video track in remote stream.')
        if not options.audio and not options.video:
            raise ValueError('Cannot subscribe a stream without audio and video.')
        if options.audio and isinstance(options.audio, AudioSubscriptionConstraints):
            if not options.audio.codecs:
                raise ValueError('Audio codec cannot be none or an empty array.')
        if options.video and isinstance(options.video, VideoSubscriptionConstraints):
            if not options.video.codecs:
                raise ValueError('Video codec cannot be none or an empty array.')
        internalId = self._createInternalId()
        await self._chainSdpPromise(internalId)
        transceivers: list[Transceiver] = []
        if options.video:
            from_id = stream.id
            parameters = None
            if isinstance(options.video, VideoSubscriptionConstraints):
                if not options.video.rid:
                    if options.video.resolution or options.video.frameRate or (options.video.bitrateMultiplier and options.video.bitrateMultiplier != 1) or options.video.keyFrameInterval:
                        parameters = VideoParameters(
                            resolution=options.video.resolution,
                            framerate=options.video.frameRate,
                            bitrate=f'x{options.video.bitrateMultiplier}' if options.video.bitrateMultiplier else None,
                            keyFrameInterval=options.video.keyFrameInterval,
                        )
                else:
                    assert stream.settings and stream.settings.video
                    matched = next((v for v in stream.settings.video if v.rid == options.video.rid), None)
                    if matched and matched._trackId:
                        from_id = matched._trackId
                        parameters = None
            transceiver = self.pc.addTransceiver('video', direction='recvonly')
            transceivers.append(VideoTransceiver(
                type='video',
                transceiver=transceiver,
                from_id=from_id,
                parameters=parameters,
                option=options.video,
            ))
        # audio must after video, or pc connect hang.
        if options.audio:
            transceiver = self.pc.addTransceiver('audio', direction='recvonly')
            transceivers.append(AudioTransceiver(
                type='audio',
                transceiver=transceiver,
                from_id=stream.id,
                option=options.audio,
            ))
        assert internalId not in self._subscribeTransceivers
        self._subscribeTransceivers[internalId] = TransceiversWithId(transceivers=transceivers)
        future: asyncio.Future[Subscription] = asyncio.Future(loop=self._loop)
        self._subscribePromises[internalId] = future
        subscriptionId = None
        try:
            await self.pc.setLocalDescription(await self.pc.createOffer())
            sub_res = await self._signaling.subscribe(remove_none_value({
                'media': {
                    'tracks': [{
                        'type': t.type,
                        'mid': t.transceiver.mid,
                        'from': t.from_id,
                        'parameters': t.parameters.to_dict() if t.parameters else None,
                    } for t in transceivers]
                },
                'transport': {
                    'id': self._id,
                    'type': 'webrtc',
                }
            }, recursion=True))
            subscriptionId = sub_res['id']
            self._subscribeTransceivers[internalId].id = subscriptionId
            self._reverseIdMap[subscriptionId] = internalId
            if self._id and self._id != sub_res['transportId']:
                raise ConferenceError(f'Server returns conflict transport ID: {sub_res["transportId"]}')
            self._id = sub_res['transportId']
            await self.dispatchEventAwaitable(MessageEvent('id', message=subscriptionId, origin=self._remoteId))
            if options:
                for t in transceivers:
                    self.pc.localDescription.sdp = self._setRtpReceiverOptions(self.pc.localDescription.sdp, options, mid=t.transceiver.mid)
            await self._signaling.soac(
                self._id,
                {
                    'type': cast(Literal['offer'], self.pc.localDescription.type),
                    'sdp': self.pc.localDescription.sdp,
                },
            )
            return await future
        except:
            await self._unsubscribe(internalId)
            raise
        
    def _onRemoteTrackAdded(self, track: MediaStreamTrack):
        logger.debug('Remote stream added.')
        for internalId, sub in self._subscribeTransceivers.items():
            if next((t for t in sub.transceivers if t.transceiver.receiver.track == track), None):
                assert sub.id is not None
                if sub.id in self._subscriptions:
                    subscription = self._subscriptions[sub.id]
                    subscription.tracks.append(track)
                    if subscription.tracks_ready() and internalId in self._subscribePromises:
                        self._subscribePromises[internalId].set_result(subscription)
                        del self._subscribePromises[internalId]
                else:
                    if sub.id not in self._remoteMediaTracks:
                        tracks = self._remoteMediaTracks[sub.id] = []
                    else:
                        tracks = self._remoteMediaTracks[sub.id]
                    tracks.append(track)
            return
            # This is not expected path. However, this is going to happen on Safari
            # because it does not support setting direction of transceiver.
        logger.warning('Received remote stream without subscription.')
        
    async def _onIceConnectionStateChange(self):
        logger.debug(f'ICE connection state changed to {self.pc.iceConnectionState}')
        async with self.exception_collector.collect():
            if self.pc.iceConnectionState == 'closed' or self.pc.iceConnectionState == 'failed':
                if self.pc.iceConnectionState == 'failed':
                    await self._handleError('connection failed.')
                else:
                    # Fire ended event if publication or subscription exists.
                    await self.close()
                
    async def _onConnectionStateChange(self):
        logger.debug(f'Connection state changed to {self.pc.connectionState}')
        async with self.exception_collector.collect():
            if self.pc.connectionState == 'closed' or self.pc.connectionState == 'failed':
                if self.pc.connectionState == 'failed':
                    await self._handleError('connection failed.')
                else:
                    # Fire ended event if publication or subscription exists.
                    await self.close()
        
    async def close_peer(self):
        if self.pc and self.pc.signalingState != 'closed':
            await self.pc.close()
        
    async def _chainSdpPromise(self, internalId: int):
        prior = self._sdpPromise
        negotiationTimeout = 10
        future = asyncio.Future(loop=self._loop)
        async def prepare_sdpResolver():
            self._sdpResolvers.append(future)
            assert internalId not in self._sdpResolverMap
            self._sdpResolverMap[internalId] = future
            return await future
        async def then(_):
            try:
                await asyncio.wait_for(prepare_sdpResolver(), timeout=negotiationTimeout)
            except asyncio.TimeoutError:
                raise ConferenceError('Timeout to get SDP answer')
        self._sdpPromise = chain_future(prior, then)
        return await chain_future(prior, catch=lambda _: None)
    
    def _nextSdpPromise(self):
        ret = False
        while self._sdpResolveNum < len(self._sdpResolvers):
            resolver = self._sdpResolvers[self._sdpResolveNum]
            self._sdpResolveNum += 1
            if not resolver.done():
                resolver.set_result(None)
                ret = True
                
        return ret
    
    def _createInternalId(self):
        id = self._internalCount
        self._internalCount += 1
        return id
        
    async def _unpublish(self, internalId: int):
        if internalId in self._publishTransceivers:
            transceivers_with_id = self._publishTransceivers[internalId]
            id = transceivers_with_id.id
            transceivers = transceivers_with_id.transceivers
            if id is not None:
                try:
                    await self._signaling.unpublish(id)
                except Exception as e:
                    logger.warning(f'MCU returns negative ack for unpublishing, {e}')
                if id in self._reverseIdMap:
                    del self._reverseIdMap[id]
            for transceiver in transceivers:
                if self.pc.signalingState == 'stable':
                    transceiver.transceiver.sender.replaceTrack(None)
                    # TODO: remove the transceiver
            del self._publishTransceivers[internalId]
            if id in self._publications:
                event = OwtEvent('ended')
                await self._publications[id].dispatchEventAwaitable(event)
                del self._publications[id]
            else:
                logger.warning(f'Invalid publication to unpublish: {id}')
                future = self._publishPromises.get(internalId)
                if future is not None:
                    if not future.done():
                        future.set_exception(ConferenceError('Failed to publish'))
                    del self._publishPromises[internalId]
            future = self._sdpResolverMap.get(internalId)
            if future is not None:
                if not future.done():
                    future.set_result(None)
                del self._sdpResolverMap[internalId]
                # not del from _sdpResolvers here
            
        
    async def _unsubscribe(self, internalId: int):
        if internalId in self._subscribeTransceivers:
            transceivers_with_id = self._subscribeTransceivers[internalId]
            id = transceivers_with_id.id
            transceivers = transceivers_with_id.transceivers
            if id is not None:
                try:
                    await self._signaling.unsubscribe(id)
                except Exception as e:
                    logger.warning(f'MCU returns negative ack for unsubscribing, {e}')
                if id in self._reverseIdMap:
                    del self._reverseIdMap[id]
            for transceiver in transceivers:
                if transceiver.transceiver.receiver and transceiver.transceiver.receiver.track:
                    transceiver.transceiver.receiver.track.stop()
            del self._subscribeTransceivers[internalId]
            if id in self._subscriptions:
                event = OwtEvent('ended')
                await self._subscriptions[id].dispatchEventAwaitable(event)
                del self._subscriptions[id]
            else:
                logger.warning(f'Invalid subscription to unsubscribe: {id}')
                future = self._subscribePromises.get(internalId)
                if future is not None:
                    if not future.done():
                        future.set_exception(ConferenceError('Failed to subscribe'))
                    del self._subscribePromises[internalId]
            future = self._sdpResolverMap.get(internalId)
            if future is not None:
                if not future.done():
                    future.set_result(None)
                del self._sdpResolverMap[internalId]
                
    async def _muteOrUnmute(self, sessionId: str, isMute: bool, isPub: bool, trackKind: TrackKink):
        event_name = 'stream-control' if isPub else 'subscription-control'
        operation = 'pause' if isMute else 'play'
        await self._signaling.ready()
        await self._signaling.send(event_name, { 'id': sessionId, 'operation': operation, 'data': trackKind })
        if not isPub:
            mute_event_name = 'mute' if isMute else 'unmute'
            subscription = self._subscriptions.get(sessionId)
            assert subscription is not None
            await subscription.dispatchEventAwaitable(MuteEvent(mute_event_name, trackKind))
                      
        
    def _readyHandler(self, sessionId: str):
        internalId = self._reverseIdMap.get(sessionId)
        if internalId in self._subscribePromises:
            try:
                assert internalId in self._subscribeTransceivers
                tracks = self._remoteMediaTracks.get(sessionId, [])
                transceivers = self._subscribeTransceivers[internalId].transceivers
                assert self._id is not None
                transportSettings = TransportSettings('webrtc', self._id, transceivers)
                subscription = Subscription(
                    sessionId,
                    tracks,
                    None,
                    transportSettings,
                    lambda: self._unsubscribe(internalId),
                    lambda kind: self._muteOrUnmute(sessionId, True, False, kind),
                    lambda kind: self._muteOrUnmute(sessionId, False, False, kind),
                )
                assert sessionId not in self._subscriptions
                self._subscriptions[sessionId] = subscription
                del self._remoteMediaTracks[sessionId]
                if subscription.tracks_ready():
                    self._subscribePromises[internalId].set_result(subscription)
                    del self._subscribePromises[internalId]
            except BaseException as e:
                self._subscribePromises[internalId].set_exception(e)
                del self._subscribePromises[internalId]
        elif internalId in self._publishPromises:
            try:
                assert internalId in self._publishTransceivers
                transceivers = self._publishTransceivers[internalId].transceivers
                assert self._id is not None
                transportSettings = TransportSettings('webrtc', self._id, transceivers)
                publication = Publication(
                    sessionId,
                    transportSettings,
                    lambda: self._unpublish(internalId),
                    lambda kind: self._muteOrUnmute(sessionId, True, True, kind),
                    lambda kind: self._muteOrUnmute(sessionId, False, True, kind),
                )
                assert sessionId not in self._publications
                self._publications[sessionId] = publication
                self._publishPromises[internalId].set_result(publication)
            except BaseException as e:
                self._publishPromises[internalId].set_exception(e)
            # Do not fire publication's ended event when associated stream is ended.
            # It may still sending silence or black frames.
            # Refer to https://w3c.github.io/webrtc-pc/#rtcrtpsender-interface.
        elif not sessionId:
            # Channel ready
            pass
        
        
    async def _sdpHandler(self, sdp: OfferAnswer | CandidateMessage):
        if sdp['type'] == 'answer':
            sdp_sdp = sdp.get('sdp')
            assert sdp_sdp is not None
            _sdp = RTCSessionDescription(sdp_sdp, sdp['type'])
            try:
                logger.debug(sdp_sdp)
                await self.pc.setRemoteDescription(_sdp)
            except Exception as e:
                logger.error(f'Set remote description failed: {e}')
                await self._rejectPromise(e)
                await self.close()
            if not self._nextSdpPromise():
                logger.warning('Unexpected SDP promise state')
                
    async def _errorHandler(self, sessionId: str | None, errorMessage: str):
        if not sessionId:
            return await self._handleError(errorMessage)
        errorEvent = ErrorEvent('error', ConferenceError(errorMessage))
        if sessionId in self._publications:
            await self._publications[sessionId].dispatchEventAwaitable(errorEvent)
        if sessionId in self._subscriptions:
            await self._subscriptions[sessionId].dispatchEventAwaitable(errorEvent)
        internalId = self._reverseIdMap.get(sessionId)
        if internalId in self._publishTransceivers:
            await self._unpublish(internalId)
        if internalId in self._subscribeTransceivers:
            await self._unsubscribe(internalId)
        
            
    async def _handleError(self, errorMessage: str):
        if self._ended:
            return
        error = ConferenceError(errorMessage)
        event = ErrorEvent('error', error=error)
        for publication in self._publications.values():
            await publication.dispatchEventAwaitable(event)
        for subscription in self._subscriptions.values():
            await subscription.dispatchEventAwaitable(event)
        await self.close()
        
    def _getAudioCodecs(self, options: PublishOptions | SubscribeOptions | SubscriptionUpdateOptions) -> list[AudioCodec]:
        if isinstance(options, PublishOptions):
            audio = options.audio
            if PublishOptions.isAudioEncodingParameters(audio): 
                return [p.codec.name for p in audio if p.codec is not None]
        elif isinstance(options, SubscribeOptions):
            audio = options.audio
            if audio and not isinstance(audio, bool) and audio.codecs is not None:
                return [codec.name for codec in audio.codecs]
        return []
    
    def _getVideoCodecs(self, options: PublishOptions | SubscribeOptions | SubscriptionUpdateOptions) -> list[VideoCodec]:
        if isinstance(options, PublishOptions):
            video = options.video
            if PublishOptions.isVideoEncodingParameters(video):
                return [p.codec.name for p in video if p.codec is not None]
        elif isinstance(options, SubscribeOptions):
            video = options.video
            if video and not isinstance(video, bool) and video.codecs is not None:
                return [codec.name for codec in video.codecs]
        return []
        
    def _setCodecOrder(self, sdp: str, options: PublishOptions | SubscribeOptions | SubscriptionUpdateOptions, mid: str | None = None) -> str:
        codecs = self._getAudioCodecs(options)
        if codecs:
            sdp = reorderCodecs(sdp, 'audio', cast(list[str], codecs), mid)
        codecs = self._getVideoCodecs(options)
        if codecs:
            sdp = reorderCodecs(sdp, 'video', cast(list[str], codecs), mid)
        return sdp
    
    def _setMaxBitrate(self, sdp: str, options: PublishOptions | SubscribeOptions | SubscriptionUpdateOptions, mid: str | None = None) -> str:
        if not isinstance(options, SubscriptionUpdateOptions) and isinstance(options.audio, list) and PublishOptions.isAudioEncodingParameters(options.audio):
            sdp = setMaxBitrate(sdp, options.audio, mid)
        if isinstance(options.video, list) and PublishOptions.isVideoEncodingParameters(options.video):
            sdp = setMaxBitrate(sdp, options.video, mid)
        return sdp
            
    def _setRtpSenderOptions(self, sdp: str, options: PublishOptions | SubscribeOptions | SubscriptionUpdateOptions, mid: str | None = None) -> str:
        return self._setMaxBitrate(sdp, options, mid)
    
    def _setRtpReceiverOptions(self, sdp: str, options: PublishOptions | SubscribeOptions | SubscriptionUpdateOptions, video_codecs: list[str] | None = None, mid: str | None = None) -> str:
        if isinstance(options, PublishOptions):
            if PublishOptions.isRTCRtpEncodingParameters(options.video) and video_codecs:
                sdp = reorderCodecs(sdp, 'video', video_codecs, mid)
                return sdp
            if PublishOptions.isRTCRtpEncodingParameters(options.audio) or PublishOptions.isRTCRtpEncodingParameters(options.video):
                return sdp
        sdp = self._setCodecOrder(sdp, options, mid)
        return sdp    
        
    @Postponer.enchance
    async def close(self):
        if self._ended:
            return
        async with self.exception_collector.collect():
            self._ended = True
            publications = list(self._publications.values())
            for publication in publications:
                await publication.stop()
            subscriptions = list(self._subscriptions.values())
            for subscription in subscriptions:
                await subscription.stop()
            await self.dispatchEventAwaitable(OwtEvent('ended'))
            await self.close_peer()
            self.exception_collector.complete()
        
    async def _rejectPromise(self, error: Exception | None):
        if not error:
            error = ConferenceError('Connection failed or closed.')
        for promise in self._publishPromises.values():
            promise.set_exception(error)
        self._publishPromises = {}
        for promise in self._subscribePromises.values():
            promise.set_exception(error)
        self._subscribePromises = {}
        
    async def _onStreamEvent(self, message: StreamUpdateMessage):
        eventTargets: list[EventDispatcher] = []
        msg_id = message['id']
        if msg_id in self._publications:
            eventTargets.append(self._publications[msg_id])
        for subscription in self._subscriptions.values():
            # 
            if msg_id == subscription._audioTrackId or msg_id == subscription._videoTrackId:
                eventTargets.append(subscription)
        if len(eventTargets) == 0:
            logger.warning('this should not happen.')
            return
        if message['status'] != 'update':
            return
        if message['data']['field'] == 'audio.status':
            trackKind: TrackKink = 'audio'
        elif message['data']['field'] == 'video.status':
            trackKind: TrackKink = 'video'
        else:
            logger.warning('Invalid data field for stream update info.')
            return
        if message['data']['value'] == 'active':
            for target in eventTargets:
                await target.dispatchEventAwaitable(MuteEvent('unmute', trackKind))
        elif message['data']['value'] == 'inactive':
            for target in eventTargets:
                await target.dispatchEventAwaitable(MuteEvent('mute', trackKind))
        


from typing import Literal

AudioCodec = Literal['pcmu', 'pcma', 'opus', 'g722', 'iSAC', 'iLBC', 'aac', 'ac3', 'nellymoser']
VideoCodec = Literal['vp8', 'vp9', 'h264', 'h265', 'av1', 'av1x']
VideoProfile = Literal['B', 'CB', 'M', 'E', 'H']

AudioSourceInfo = Literal['mic', 'screen-cast', 'raw-file', 'encoded-file', 'mixed']
"""
Source info about an audio track. Values: 'mic', 'screen-cast',
'file', 'mixed'.
"""

VideoSourceInfo = Literal['camera', 'screen-cast', 'raw-file', 'encoded-file', 'mixed']
"""
Source info about a video track. Values: 'camera', 'screen-cast',
'file', 'mixed'.
"""

TransportType = Literal['quic', 'webrtc']


import asyncio
from aioquic.asyncio.client import QuicConfiguration
import base64
import json
import logging
import sys
from dataclasses import dataclass, field
from enum import Enum
from threading import Lock
from typing import Callable, cast, overload

from aiortc import RTCConfiguration, RTCIceServer, RTCRtpTransceiver

if sys.version_info < (3, 11):
    from typing_extensions import Any, Literal, NotRequired, TypedDict
else:
    from typing import Any, Literal, NotRequired, TypedDict
    
if __name__ == '__main__':
    from pathlib import Path
    file = Path(__file__).resolve()
    parent, top = file.parent, file.parents[1]
    sys.path.append(str(top))
    try:
        sys.path.remove(str(parent))
    except ValueError: # already removed
        pass
    __package__ = '.'.join(parent.parts[len(top.parts):])

from .channel import OwtChannel
from .quic import QuicChannel
from .common import ConferenceError, EventDispatcher, MessageEvent, OwtEvent
from .constants import AudioSourceInfo, VideoSourceInfo
from .info import ConferenceInfo
from .participant import Participant
from .portal_types import (LoginInfo, ParticipantAction,
                           StreamAddUpdateMessage, StreamInfo,
                           StreamRemoveUpdateMessage, StreamUpdateMessage,
                           StreamUpdateUpdateMessage, TextReceiveMessage)
from .publication import PublishOptions
from .signaling import SioSignaling
from .stream import (ActiveAudioInputChangeEvent, LayoutChangeEvent,
                     LocalStream, RemoteMixedStream, RemoteStream, StreamEvent,
                     StreamSourceInfo, convertToPublicationSettings,
                     convertToSubscriptionCapabilities)
from .subscription import SubscribeOptions
from .utils import (AsyncExceptionCollector, NullAsyncContext, ensure_future)

logger = logging.getLogger('owt[client]')

class SignalingState(Enum):
    READY = 1
    CONNECTING = 2
    CONNECTED = 3
    


class Token(TypedDict):
    host: str
    secure: NotRequired[bool | None]
    webTransportUrl: NotRequired[str | None]
    

@dataclass
class WebTransportHash:
    algorithm: str
    value: bytes
    
@dataclass
class WebTransportOptions:
    allowPooling: bool = False
    requireUnreliable: bool = False
    serverCertificateHashes: list[WebTransportHash] = field(default_factory=list)
    congestionControl: Literal['default', 'hroughput', 'low-latency'] = 'default'
    
class ParticipantEvent(OwtEvent):
    participant: Participant
    
    def __init__(self, type: str, participant: Participant) -> None:
        super().__init__(type)
        self.participant = participant

@dataclass
class ConferenceClientConfiguration:
    rtcConfiguration: RTCConfiguration | None = None
    webTransportConfiguration: QuicConfiguration | None = None

class ConferenceClient(EventDispatcher):
    config: ConferenceClientConfiguration
    signalingState: SignalingState = SignalingState.READY
    signaling: SioSignaling
    me: Participant | None
    __ready: asyncio.Event
    __ready_lock: asyncio.Lock
    __loop: asyncio.AbstractEventLoop | None
    __remoteStreams: dict[str, RemoteStream]
    __participants: dict[str, Participant]
    __channels: dict[str, OwtChannel]
    quicTransportChannel: QuicChannel | None
    exception_collector: AsyncExceptionCollector | None
    
    def __init__(self, config: ConferenceClientConfiguration | None = None, loop: asyncio.AbstractEventLoop | None = None) -> None:
        super().__init__()
        self.config = config if config is not None else ConferenceClientConfiguration()
        self.me = None
        self.__loop = loop
        self.__ready = asyncio.Event()
        self.__ready_lock = asyncio.Lock()
        self.signaling = SioSignaling(loop=loop)
        self.__remoteStreams = {}
        self.__participants = {}
        self.__channels = {}
        self.quicTransportChannel = None
        self.exception_collector = None
        self.signaling.addEventListener('data', self.onDataMessage)
        self.signaling.addEventListener('disconnect', self.onDisconnect)
        
        
    @property
    def loop(self):
        return self.__loop
    
    @property
    def remoteStreams(self):
        return self.__remoteStreams
    
    @property
    def participants(self):
        return self.__participants
    
    @property
    def channels(self):
        return self.__channels
    
    async def onDisconnect(self, _):
        await self.leave()
        await self.dispatchEventAwaitable(OwtEvent('serverdisconnected'))
    
    async def onDataMessage(self, evt: MessageEvent):
        await self.onSignalingMessage(notification=evt.message['notification'], data=evt.message['data'])
    
    async def onSignalingMessage(self, notification: str, data: Any):
        if notification == 'soac' or notification == 'progress':
            id = data.get('id')
            if id in self.channels:
                await self.channels[id].onMessage(notification, data)
            elif False: # TODO: quicTransportChannel part
                pass
            else:
                logger.warning(f'Cannot find a channel for incoming data: {id}.')
        elif notification == 'stream':
            stream_data = cast(StreamUpdateMessage, data)
            status = stream_data['status']
            if status == 'add':
                stream_data = cast(StreamAddUpdateMessage, stream_data)
                await self.fireStreamAdded(stream_data['data'])
            elif status == 'remove':
                stream_data = cast(StreamRemoveUpdateMessage, stream_data)
                await self.fireStreamRemoved(stream_data)
            elif status == 'update':
                stream_data = cast(StreamUpdateUpdateMessage, stream_data)
                if stream_data['data']['field'] == 'audio.status' or stream_data['data']['field'] == 'video.status':
                    for channel in self.channels.values():
                        await channel.onMessage(notification, stream_data)
                elif stream_data['data']['field'] == 'activeInput':
                    await self.fireActiveAudioInputChange(stream_data)
                elif stream_data['data']['field'] == 'video.layout':
                    await self.fireLayoutChange(stream_data)
                elif stream_data['data']['field'] == '.':
                    await self.updateRemoteStream(stream_data['data']['value'])
                else:
                    raise ConferenceError('not support yet.')
            else:
                raise ConferenceError('not support yet.')
        elif notification == 'text':
            await self.fireMessageReceived(data)
        elif notification == 'participant':
            await self.fireParticipantEvent(data)
        else:
            logger.warning(f'Unknown notification: {notification}')
            
    async def updateRemoteStream(self, stream_info: StreamInfo):
        if stream_info['id'] not in self.remoteStreams:
            logger.warning(f'Cannot find specific remote stream: {stream_info["id"]}')
            return
        stream = self.remoteStreams[stream_info['id']]
        stream.settings = convertToPublicationSettings(stream_info.get('media'))
        stream.extraCapabilities = convertToSubscriptionCapabilities(stream_info.get('media'))
        await stream.dispatchEventAwaitable(OwtEvent('updated'))
    
    def _createRemoteStream(self, stream_info: StreamInfo) -> RemoteStream:
        if stream_info['type'] == 'mixed':
            return RemoteMixedStream(stream_info)
        media = stream_info.get('media')
        audio_track = next((t for t in (media['tracks'] if media else []) if t['type'] == 'audio'), None)
        video_track = next((t for t in (media['tracks'] if media else []) if t['type'] == 'video'), None)
        info = stream_info.get('info')
        owner = info.get('owner')
        stream = RemoteStream(
            tracks=[],
            quic_stream=None,
            id=stream_info.get('id'),
            origin=owner,
            source=StreamSourceInfo(
                audio=cast(AudioSourceInfo | None, audio_track.get('source') if audio_track is not None else None),
                video=cast(VideoSourceInfo | None, video_track.get('source') if video_track is not None else None),
                data=stream_info.get('data', False),
            ),
            attributes=info.get('attributes') or {},
        )
        stream.settings = convertToPublicationSettings(media)
        stream.extraCapabilities = convertToSubscriptionCapabilities(media)
        return stream
    
    def createPeerConnectionChannel(self) -> OwtChannel:
        channel = OwtChannel(signaling=self.signaling, rtc_configure=self.config.rtcConfiguration, loop=self.__loop)
        def on_id_msg(evt: MessageEvent):
            if evt.message not in self.channels:
                self.channels[evt.message] = channel
            else:
                logger.warning(f'Channel already exists {evt.message}')
        def on_ended(_):
            id = next((k for k, c in self.channels.items() if c == channel), None)
            if id is None:
                logger.warning('Receive ended message from unknown channel.')
            else:
                del self.channels[id]
        channel.addEventListener('id', on_id_msg)
        channel.addEventListener('ended', on_ended)
        return channel
    
    def collector_context(self):
        collector = self.exception_collector
        return collector.collect() if collector else NullAsyncContext()
    
    async def join(self, tokenString: str):
        async with self.__ready_lock:
            if self.__ready.is_set():
                raise ConferenceError('The client has joined the conference, to join another conference, leave at first.')
            token: Token = json.loads(base64.b64decode(tokenString))
            isSecured = token.get('secure', False)
            host = token['host']
            if host.find('http') == -1:
                host = f'https://{host}' if isSecured else f'http://{host}'
            if self.signalingState != SignalingState.READY:
                raise ConferenceError('connection state invalid')
            self.signalingState = SignalingState.CONNECTING
            self.exception_collector = AsyncExceptionCollector(loop=self.__loop)
            try:
                login_info: LoginInfo = {
                    'token': tokenString,
                    'userAgent': {
                        'sdk': {
                            'version': '5.0',
                            'type': 'JavaScript',
                        }
                    },
                    'protocol': '1.2'
                }
                login_result = await self.signaling.connect(host, login_info)
                self.signalingState = SignalingState.CONNECTED
                room = login_result['room']
                streams = room.get('streams')
                if streams:
                    for st in streams:
                        self.remoteStreams[st['id']] = self._createRemoteStream(st)
                participants = room.get('participants')
                if participants:
                    for p in participants:
                        self.participants[p['id']] = Participant(p['id'], p['role'], p['user'])
                        if p['id'] == login_result['id']:
                            self.me = self.participants[p['id']]
                quic_url = token.get('webTransportUrl')
                if quic_url:
                    quic_token = login_result.get('webTransportToken')
                    if not quic_token:
                        logger.warning('Unable to find quic token in login result, but the login token support it.')
                    else:
                        self.quicTransportChannel = QuicChannel(quic_url, quic_token, self.signaling, configure=self.config.webTransportConfiguration)
                        await self.quicTransportChannel.init()
                assert self.me is not None
                conferenceInfo = ConferenceInfo(
                    id=room['id'],
                    participants=list(self.participants.values()),
                    remoteStreams=list(self.remoteStreams.values()),
                    self=self.me,
                )
                self.__ready.set()
                return conferenceInfo
            except Exception as e:
                await self._try_leave()
                raise e
        
    async def _try_leave(self):
        if self.__ready.is_set():
            self.__ready.clear()
        channels = list(self.channels.values())
        for channel in channels:
            await channel.close()
        self.channels.clear()
        if self.quicTransportChannel:
            await self.quicTransportChannel.close()
            self.quicTransportChannel = None
        await self.signaling.disconnect(permit_duplicate_disconnect=True)
        self.clean()
        self.signalingState = SignalingState.READY
        self.exception_collector = None
        
    async def leave(self, may_be_leaved: bool = True):
        if not self.__ready.is_set():
            if may_be_leaved:
                return
            else:
                raise asyncio.InvalidStateError('The client has leaved the conference.')
        async with self.__ready_lock:
            if not self.__ready.is_set():
                if may_be_leaved:
                    return
                else:
                    raise asyncio.InvalidStateError('The client has leaved the conference.')
            await self._try_leave()
            
    def joined(self):
        return self.__ready.is_set()
    
    async def wait_joined(self):
        return await self.__ready.wait()
        
    async def publish(self, stream: LocalStream, options: PublishOptions | list[RTCRtpTransceiver] | None = None, video_codecs: list[str] | None = None):
        if not self.joined():
            raise asyncio.InvalidStateError('Join a conference at first.')
        if stream.source.data:
            if not self.quicTransportChannel:
                raise ConferenceError('The quic channel is not support, so unable to publish data stream.')
            return await self.quicTransportChannel.publish(stream)
        channel = self.createPeerConnectionChannel()
        pub = await channel.publish(stream, options, video_codecs)
        async def on_ended(_):
            await channel.close()
        pub.addEventListener('ended', on_ended)
        return pub
    
    async def subscribe(self, stream: RemoteStream, options: SubscribeOptions | None = None):
        if not self.joined():
            raise asyncio.InvalidStateError('Join a conference at first.')
        if stream.source.data:
            if stream.source.audio or stream.source.video:
                raise ValueError('Invalid source info. A remote stream is either a data stream or a media stream.')
            if not self.quicTransportChannel:
                raise ConferenceError('The quic channel is not support, so unable to subscribe data stream.')
            return await self.quicTransportChannel.subscribe(stream)
        else:
            channel = self.createPeerConnectionChannel()
            sub = await channel.subscribe(stream, options)
            async def on_ended(_):
                await channel.close()
            sub.addEventListener('ended', on_ended)
            return sub
        
    async def wait_for_stream(self, filter: Callable[[RemoteStream], bool], timeout: float = 0):
        stream_event = asyncio.Event()
        def on_stream_add(evt: StreamEvent):
            if filter(evt.stream):
                stream_event.set()
                self.removeEventListener('streamadded', on_stream_add)
        self.addEventListener('streamadded', on_stream_add)
        stream = next((st for st in self.remoteStreams.values() if filter(st)), None)
        if stream:
            self.removeEventListener('streamadded', on_stream_add)
            return stream
        else:
            if timeout > 0:
                await asyncio.wait_for(stream_event.wait(), timeout=timeout)
            else:
                await stream_event.wait()
            stream = next((st for st in self.remoteStreams.values() if filter(st)), None)
            assert stream
            return stream
        
    def consume_streams(self, callback: Callable[[RemoteStream], Any], auto_resume: bool = True, timeout: float = 0):
        loop = self.__loop
        class Tasks:
            tasks: list[asyncio.Future]
            lock: Lock
            state: Literal['running', 'timeouting', 'timeouted', 'error', 'closing', 'closed']
            timeout: float
            working_future: asyncio.Future
            timeout_future: asyncio.Future
            
            def __init__(self, timeout: float) -> None:
                self.tasks = []
                self.lock = Lock()
                self.working_future = asyncio.Future()
                self.working_future.set_result(None)
                self.state = 'timeouting'
                self.timeout_future = asyncio.ensure_future(self._do_timeout(), loop=loop)
                self.timeout = timeout
                    
                    
            async def _do_timeout(self):
                if self.timeout > 0:
                    await asyncio.sleep(self.timeout)
                with self.lock:
                    if self.state == 'timeouting':
                        self.state = 'timeouted'
                        
                
            async def wait(self, close: bool = True) -> None:
                while True:
                    try:
                        await self.working_future
                        await self.timeout_future
                    except asyncio.CancelledError:
                        if self.state == 'closing' or self.state == 'closed':
                            break
                    except BaseException:
                        self.close()
                        raise
                    else:
                        with self.lock:
                            if self.state == 'timeouted':
                                self.state = 'closing'
                                break
                            elif self.state == 'closed' or self.state == 'closing':
                                break
                if close:
                    self.close()
                    
            def renew(self) -> bool:
                if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                    return False
                with self.lock:
                    if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                        return False
                    elif self.state == 'running':
                        return True
                    elif self.state == 'timeouting' or self.state == 'timeouted':
                        self.timeout_future = asyncio.ensure_future(self._do_timeout(), loop=loop)
                        self.state = 'timeouting'
                        return True
                    else:
                        raise asyncio.InvalidStateError(f'Invalid state: {self.state}')
                
            def _on_task_done(self, task: asyncio.Future):
                if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                    return
                if task in self.tasks:
                    with self.lock:
                        if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                            return
                        if task in self.tasks:
                            self.tasks.remove(task)
                            if not self.tasks:
                                assert not self.working_future.done()
                                self.working_future.set_result(None)
                                assert self.timeout_future.done()
                                self.timeout_future = asyncio.ensure_future(self._do_timeout(), loop=loop)
                                self.state = 'timeouting'
                
            def add_task(self, task: asyncio.Future):
                if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                    task.cancel()
                    return
                with self.lock:
                    if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                        task.cancel()
                    else:
                        if not self.tasks:
                            assert self.state != 'running'
                            assert self.working_future.done()
                            self.state = 'running'
                            self.working_future = asyncio.Future()
                            self.timeout_future.cancel()
                        task.add_done_callback(self._on_task_done)
                        self.tasks.append(task)
                        
            def set_exception(self, e: BaseException):
                if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                    return
                with self.lock:
                    if self.state == 'closed' or self.state == 'closing' or self.state == 'error':
                        return
                    self.state = 'error'
                    for task in self.tasks:
                        task.cancel()
                    self.working_future.cancel()
                    self.working_future = asyncio.Future()
                    self.working_future.set_exception(e)
                    self.timeout_future.cancel()
                    self.timeout_future = asyncio.Future()
                    self.timeout_future.set_exception(e)
                    
                        
            def close(self):
                if self.state == 'closed':
                    return
                with self.lock:
                    if self.state == 'closed':
                        return
                    else:
                        self.state = 'closed'
                        for task in self.tasks:
                            task.cancel()
                        self.tasks.clear()
                        self.working_future.cancel()
                        self.timeout_future.cancel()
                        return
            
        class Handle:
            tasks: Tasks
            client: ConferenceClient
            listener: Callable
            closed: bool
            
            def __init__(self, tasks: Tasks, client: ConferenceClient, listener: Callable) -> None:
                self.tasks = tasks
                self.client = client
                self.listener = listener
                self.closed = False
                self.client.signaling.addEventListener('connected', self._on_sig_connected)
                asyncio.ensure_future(self._on_sig_connected(None))
                
            async def _on_sig_connected(self, _):
                if self.closed:
                    return
                collector = self.client.signaling.exception_collector
                if collector:
                    await collector.add_listener(self._on_exception)
                
            def close(self):
                self.closed = True
                self.client.removeEventListener('streamadded', self.listener)
                self.tasks.close()
                self.client.signaling.removeEventListener('connected', self._on_sig_connected)
                collector = self.client.signaling.exception_collector
                if collector:
                    collector.remove_listener(self._on_exception)
                
            async def wait(self):
                try:
                    await self.tasks.wait()
                finally:
                    self.close()
                
            def renew(self) -> bool:
                return tasks.renew()
            
            def _on_exception(self, e: BaseException):
                self.tasks.set_exception(e)
            
            async def link_exception_collector(self, collector: AsyncExceptionCollector):
                await collector.add_listener(self._on_exception)
                
        tasks: Tasks = Tasks(timeout=timeout)
        async def wrap_task(stream: RemoteStream):
            while True:
                try:
                    await ensure_future(callback, stream)
                except (asyncio.TimeoutError, asyncio.CancelledError):
                    raise
                except BaseException as e:
                    if stream.id not in self.remoteStreams:
                        logger.info(f'Stream {stream.id} removed, so stop the binding task.')
                        break
                    if not auto_resume:
                        raise
                else:
                    break
                    
        def on_stream_add(evt: StreamEvent):
            tasks.add_task(asyncio.ensure_future(wrap_task(evt.stream), loop=self.__loop))
        for stream in self.remoteStreams.values():
            tasks.add_task(asyncio.ensure_future(wrap_task(stream), loop=self.__loop))
        self.addEventListener('streamadded', on_stream_add)
        handle = Handle(tasks, self, on_stream_add)
        return handle
    
    async def send(self, message: str, participantId: str = 'all'):
        """
        Send a text message to a participant or all participants.

        Args:
            message (str): Message to be sent.
            participantId (str | None): Receiver of this message. 'all' means all participants.
        """
        await self.signaling.send_text(message, participantId)
        
    def is_support_data_stream(self):
        if not self.joined():
            raise asyncio.InvalidStateError('Join a conference at first.')
        return self.quicTransportChannel is not None
        
    async def create_data_local_stream(self, attributes: dict | None = None, is_unidirectional: bool = False) -> LocalStream:
        if not self.joined():
            raise asyncio.InvalidStateError('Join a conference at first.')
        if not self.quicTransportChannel:
            raise ConferenceError('No QUIC connection available. Perhaps the conference not support quic connection.')
        return await self.quicTransportChannel.createSendStream(attributes=attributes, is_unidirectional=is_unidirectional)
        
    def clean(self):
        self.participants.clear()
        self.remoteStreams.clear()
        
    @overload
    def addEventListener(self, eventType: Literal['streamadded'], listener: Callable[[StreamEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: Literal['messagereceived'], listener: Callable[[MessageEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    @overload
    def addEventListener(self, eventType: Literal['participantjoined'], listener: Callable[[ParticipantEvent], Any], once: bool = False, seq: bool = True) -> None: ...
    def addEventListener(self, eventType: str, listener: Callable[[Any], Any], once: bool = False, seq: bool = True) -> None:
        return super().addEventListener(eventType, listener, once=once, seq=seq)
        
    async def fireMessageReceived(self, data: TextReceiveMessage):
        await self.dispatchEventAwaitable(MessageEvent('messagereceived', message=data['message'], origin=data['from'], to=data['to']))
        
    async def fireStreamAdded(self, info: StreamInfo):
        stream = self._createRemoteStream(info)
        self.remoteStreams[stream.id] = stream
        await self.dispatchEventAwaitable(StreamEvent('streamadded', stream))
        
    async def fireStreamRemoved(self, info: StreamRemoveUpdateMessage):
        if info['id'] not in self.remoteStreams:
            logger.warning(f'Cannot find specific remote stream {info["id"]}.')
            return
        stream = self.remoteStreams[info['id']]
        del self.remoteStreams[info['id']]
        await stream.dispatchEventAwaitable(OwtEvent('ended'))
        
    async def fireActiveAudioInputChange(self, info: StreamUpdateUpdateMessage):
        if info['id'] not in self.remoteStreams:
            logger.warning(f'Cannot find specific remote stream {info["id"]}.')
            return
        stream = self.remoteStreams[info['id']]
        assert info['data']['field'] == 'activeInput'
        event = ActiveAudioInputChangeEvent('activeaudioinputchange', info['data']['value'])
        await stream.dispatchEventAwaitable(event)
        
        
    async def fireLayoutChange(self, info: StreamUpdateUpdateMessage):
        if info['id'] not in self.remoteStreams:
            logger.warning(f'Cannot find specific remote stream {info["id"]}.')
            return
        stream = self.remoteStreams[info['id']]
        event = LayoutChangeEvent('layoutchange', info['data']['value'])
        await stream.dispatchEventAwaitable(event)
        
        
    async def fireParticipantEvent(self, data: ParticipantAction):
        if data['action'] == 'join':
            participant = Participant(id=data['data']['id'], role=data['data']['role'], userId=data['data']['user'])
            self.participants[participant.id] = participant
            await self.dispatchEventAwaitable(ParticipantEvent('participantjoined', participant))
        elif data['action'] == 'leave':
            participantId = data['data']
            if participantId not in self.participants:
                logger.warning(f'Received leave message from MCU for an unknown participant: {participantId}.')
                return
            participant = self.participants[participantId]
            del self.participants[participantId]
            await participant.dispatchEventAwaitable(OwtEvent('left'))

import platform

from aiortc.contrib.media import MediaPlayer


def create_local_tracks(type: Literal['camera', 'file'], path: str | None = None):
    options = {"framerate": "20", "video_size": "320x240"}
    
    if type == 'camera':
        try:
            if platform.system() == "Darwin":
                webcam = MediaPlayer(
                    "default:none", format="avfoundation", options=options
                )
            elif platform.system() == "Windows":
                # use 'ffmpeg -list_devices true -f dshow -i dummy' to get the camera name
                from device import getDeviceList
                devices = getDeviceList()
                if not devices:
                    raise ValueError('No camera found.')
                webcam = MediaPlayer(
                    f"video={devices[0][0]}", format="dshow", options=options
                )
            else:
                webcam = MediaPlayer("/dev/video0", format="v4l2", options=options)
        except BaseException as e:
            raise ValueError(f'Unable to get a camera. {e}')
    elif type == 'file':
        assert path is not None
        webcam = MediaPlayer(path, options=options)
    else:
        raise ValueError(f'invalid type {type}')
    stream = LocalStream(
        tracks=[t for t in [webcam.audio, webcam.video] if t is not None],
        quic_stream=None,
        source=StreamSourceInfo(audio='mic', video='camera'),
    )
    return stream
    


import sys

if sys.version_info < (3, 11):
    from typing_extensions import (Any, Literal, NotRequired, TypeAlias,
                                   TypedDict, overload)
else:
    from typing import TypedDict, NotRequired, Any, Literal, overload, TypeAlias

from .constants import (AudioCodec, AudioSourceInfo, TransportType, VideoCodec,
                        VideoProfile, VideoSourceInfo)


class ReconnectionOptions(TypedDict):
    keepTime: float
    """
    -1: Use server side configured 'reconnection_timeout' value; Others: specified keepTime in seconds
    """

class Sdk(TypedDict):
    type: str
    version: str

class ClientInfo(TypedDict):
    sdk: Sdk

class LoginInfo(TypedDict):
    token: str
    userAgent: ClientInfo
    reconnection: NotRequired[ReconnectionOptions | Literal[False] | None]
    protocol: NotRequired[str | None]
    
class PendingMessage(TypedDict):
    event: str
    data: Any
    seq: int
    """
    Internal Seq No
    """
    
class ReloginResponse(TypedDict):
    ticket: str
    """
    base64 encoded ReconnectionTicket object
    """
    messages: list[PendingMessage]
    
class PermissionData(TypedDict):
    audio: bool
    video: bool
    data: bool
    
class Permission(TypedDict):
    publish: PermissionData
    subscribe: PermissionData
    
class ParticipantInfo(TypedDict):
    id: str
    """
    The ID of the participant. It varies when a single user join different conferences.
    """
    role: str
    user: str
    """
    The user ID of the participant. It can be integrated into existing account management system.
    """
    
class ParticipantJoinAction(TypedDict):
    action: Literal['join']
    data: ParticipantInfo
    
class ParticipantLeaveAction(TypedDict):
    action: Literal['leave']
    data: str
    """
    Participant Id
    """

ParticipantAction: TypeAlias = ParticipantJoinAction | ParticipantLeaveAction
    
class PublicationInfo(TypedDict):
    owner: str
    
class UnpublicationRequest(TypedDict):
    id: str
    """
    same as the stream id.
    """
    
class UnsubscriptionRequest(TypedDict):
    id: str
    """
    Subscription Id
    """
    
class StreamControlInfo(TypedDict):
    id: str
    """
    Stream id, Must refer to a forward stream
    """
    operation: Literal['pause', "play"]
    data: Literal['audio', 'video', 'av']
    """
    If operation equals "pause" or "play"
    """
    
class VideoParametersSpecification(TypedDict):
    resolution: NotRequired["Resolution | None"]
    framerate: NotRequired[float | None]
    bitrate: NotRequired[float | str | None]
    keyFrameInterval: NotRequired[float | None]
    
class AudioUpdate(TypedDict('AudioUpdate', {
    'from': str
})):
    pass
    
class VideoUpdate(TypedDict('VideoUpdate', {
    'from': NotRequired[str | None]
})):
    parameters: NotRequired[VideoParametersSpecification | None]
    
class SubscriptionUpdate(TypedDict):
    audio: NotRequired[AudioUpdate | None]
    video: NotRequired[VideoUpdate | None]
    
class SubscriptionControlInfo(TypedDict):
    id: str
    operation: Literal['update', 'pause', 'play', 'querybwe']
    data: NotRequired[SubscriptionUpdate | Literal['audio', 'video', 'av'] | None]
    """
    SubscriptionUpdate If operation equals "update",
    ("audio" | "video" | "av") If operation equals "pause" or "play",
    undefined If operation equals "querybwe".
    """
    
class BWEResult(TypedDict):
    estimatedBitrate: float
    """
    Send side estimated bitrate if enabled
    """
    

class AugmentInfo(TypedDict):
    pass

class ViewInfo(TypedDict):
    pass

class BaseStreamInfo(TypedDict):
    id: str
    media: NotRequired["MediaInfo | None"]
    data: bool

class ForwardStreamInfo(BaseStreamInfo):
    type: Literal['forward']
    info: PublicationInfo
    
class AugmentedStreamInfo(BaseStreamInfo):
    type: Literal['augmented']
    info: AugmentInfo
    
class MixedStreamInfo(BaseStreamInfo):
    type: Literal['mixed']
    info: ViewInfo
    
StreamInfo: TypeAlias = ForwardStreamInfo | AugmentedStreamInfo | MixedStreamInfo
    
class RoomInfo(TypedDict):
    id: str
    views: NotRequired[list[str] | None]
    streams: NotRequired[list[StreamInfo] | None]
    participants: NotRequired[list[ParticipantInfo] | None]
    
class LoginResult(TypedDict):
    id: str
    user: str
    role: str
    permission: Permission
    room: RoomInfo
    reconnectionTicket: NotRequired[str | None]
    webTransportToken: NotRequired[str | None]
    """
    Base64Encoded(object(ReconnectionTicket))) when reconnection is promised
    """

class ReconnectionTicket(TypedDict):
    participantId: str
    ticketId: str
    notBefore: float
    notAfter: float
    signature: str
    
class TextSendMessage(TypedDict):
    to: str
    """
    Participant Id or all
    """
    message: str
    """
    The message length must not greater than 2048
    """
    
class TextReceiveMessage(TypedDict('TextReceiveMessage', {
    'from': str, # Participant Id
})):
    to: Literal['all', 'me']
    message: str

class OfferAnswer(TypedDict):
    type: Literal['offer', 'answer']
    sdp: NotRequired[str | None]
    
class Candidate(TypedDict):
    sdpMid: str
    sdpMLineIndex: int
    candidate: str

class RemovedCandidate(TypedDict):
    sdpMid: NotRequired[str | None]
    sdpMLineIndex: NotRequired[int | None]
    candidate: str
    
class CandidateMessage(TypedDict):
    type: Literal['candidate']
    candidate: Candidate
    
class RemovedCandidatesMessage(TypedDict):
    type: Literal['removed-candidates']
    candidates: list[RemovedCandidate]
    
class SOACMessage(TypedDict):
    id: str
    """
    Transport ID returned in publishing or subscribing
    """
    signaling: OfferAnswer | CandidateMessage | RemovedCandidatesMessage
    
class RTCRtpSendParameters(TypedDict):
    pass

class RtpConfig(TypedDict):
    ssrc: str
    sendParameters: RTCRtpSendParameters # [RTCRtpSendParameters](https://www.w3.org/TR/webrtc/#rtcsendrtpparameters) defined in WebRTC 1.0.

class SoacTransportProgress(TypedDict):
    id: str # transportId
    status: Literal['soac']
    data: OfferAnswer | CandidateMessage
    
class ReadyTransportProgress(TypedDict):
    id: str # transportId
    sessionId: NotRequired[str | None]
    status: Literal['ready']
    
class ErrorTransportProgress(TypedDict):
    id: str # transportId
    status: Literal['error']
    sessionId: NotRequired[str | None]
    data: str # reason
    
class RtpTransportProgress(TypedDict):
    id: str # transportId
    status: Literal['rtp']
    data: RtpConfig
    
TransportProgress = SoacTransportProgress | ReadyTransportProgress | ErrorTransportProgress | RtpTransportProgress

class SessionProgress(TypedDict):
    id: str # StreamId returned in publishing or SubscriptionId returned in subscribing
    status: Literal['ready'] | Literal['error']

class AudioFormat(TypedDict):
    codec: NotRequired[AudioCodec]
    sampleRate: NotRequired[float]
    channelNum: NotRequired[int]
    
class VideoFormat(TypedDict):
    codec: NotRequired[VideoCodec]
    profile: NotRequired[VideoProfile]
    """
    If codec does NOT equal "h264", should be None
    """
    
class Resolution(TypedDict):
    width: int
    height: int
    
class VideoParameters(TypedDict):
    resolution: NotRequired[Resolution | None]
    framerate: NotRequired[float | None]
    bitrate: NotRequired[float | str | None]
    keyFrameInterval: NotRequired[float | None]
    
class TrackInfoOptionalParameters(TypedDict):
    resolution: NotRequired[list[Resolution] | None]
    framerate: NotRequired[list[float] | None]
    bitrate: NotRequired[list[float | str] | None]
    keyFrameInterval: NotRequired[list[float] | None]
    
class TrackInfoOptional(TypedDict):
    format: NotRequired[list[AudioFormat] | list[VideoFormat] | None]
    parameters: NotRequired[TrackInfoOptionalParameters | None]

class TrackInfo(TypedDict):
    id: NotRequired[str | None]
    type: Literal['audio', 'video']
    format: NotRequired[AudioFormat | VideoFormat | None]
    parameters: NotRequired[VideoParameters | None]
    status: NotRequired[Literal['active', 'inactive'] | None]
    source: NotRequired[AudioSourceInfo | VideoSourceInfo | None]
    mid: NotRequired[str | None]
    """
    undefined if transport's type is "quic"
    """
    rid: NotRequired[str | None]
    optional: NotRequired[TrackInfoOptional | None]
    

class MediaInfo(TypedDict):
    tracks: list[TrackInfo]
    
class TransportOptions(TypedDict):
    type: TransportType
    id: NotRequired[str | None]
    """
    null will result to create a new transport channel. Always be null if transport type is webrtc because webrtc agent doesn't support multiple transceivers on a single PeerConnection at this time.
    """

class PublicationRequest(TypedDict):
    media: NotRequired[MediaInfo | None]
    data: NotRequired[bool | None]
    transport: NotRequired[TransportOptions | None]
    attributes: NotRequired[dict | None]
    
class PublicationOrSubscriptionResult(TypedDict):
    transportId: str
    """
    Can be reused in the following publication or subscription.
    """
    id: str
    """
    will be used as the stream id when it gets ready.
    """
    
class TrackSubInfo(TypedDict('TrackSubInfo', {
    'from': NotRequired[str | None],
})):
    type: Literal['audio', 'video']
    mid: NotRequired[str | None]
    parameters: NotRequired[VideoParameters | None]
    
class MediaSubOptions(TypedDict):
    tracks: list[TrackSubInfo]
    
class MediaSubscriptionRequest(TypedDict):
    media: MediaSubOptions
    transport: NotRequired[TransportOptions | None]
    
    
class QuicSubscriptionRequestData(TypedDict('QuicSubscriptionRequestData', { "from": str })):
    pass
class QuicSubscriptionRequest(TypedDict):
    data: QuicSubscriptionRequestData
    transport: NotRequired[TransportOptions | None]
    
SubscriptionRequest: TypeAlias = MediaSubscriptionRequest | QuicSubscriptionRequest
    
class BaseStreamUpdateMessage(TypedDict):
    id: str
    """
    stream id
    """

class StreamAddUpdateMessage(BaseStreamUpdateMessage):
    status: Literal['add']
    data: StreamInfo
    
class MixedVideoLayout(TypedDict):
    pass
    
class StreamUpdateStatus(TypedDict):
    field: Literal['audio.status', 'video.status']
    value: Literal['active', 'inactive']
    
class StreamUpdateVideoLayout(TypedDict):
    field: Literal['video.layout']
    value: MixedVideoLayout
    
class StreamUpdateActiveInput(TypedDict):
    field: Literal['activeInput']
    value: str
    
class StreamUpdateWhole(TypedDict):
    field: Literal['.']
    value: StreamInfo
    
StreamUpdate: TypeAlias = StreamUpdateStatus | StreamUpdateVideoLayout | StreamUpdateActiveInput | StreamUpdateWhole

class StreamUpdateUpdateMessage(BaseStreamUpdateMessage):
    status: Literal['update']
    data: StreamUpdate
    
class StreamRemoveUpdateMessage(BaseStreamUpdateMessage):
    status: Literal['remove']
    
StreamUpdateMessage = StreamAddUpdateMessage | StreamUpdateUpdateMessage | StreamRemoveUpdateMessage